
#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_yaml.yaml.ruamel.ruamel_yaml_client import RuamelYAMLClient
from holado_yaml.yaml.pyyaml.pyyaml_client import PyYAMLClient
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_yaml.yaml.enums import UpdateType


logger = logging.getLogger(__name__)



class YAMLManager(object):
    """
    Manage actions on YAML files.
    """
    
    @classmethod
    def is_available(cls):
        return RuamelYAMLClient.is_available() or PyYAMLClient.is_available()
    
    @classmethod
    def get_client(cls, client_type, **kwargs):
        if RuamelYAMLClient.is_available():
            return RuamelYAMLClient(client_type=client_type)
        elif PyYAMLClient.is_available():
            return PyYAMLClient(client_type=client_type)
        else:
            raise TechnicalException("Missing dependencies")
    
    
    # Manage YAML files
    
    @classmethod
    def load_file(cls, file_path, client_type=None, client=None):
        client = client if client is not None else cls.get_client(client_type)
        return client.load_file(file_path)
    
    @classmethod
    def load_multiple_documents_file(cls, file_path, client_type=None, client=None):
        client = client if client is not None else cls.get_client(client_type)
        return client.load_multiple_documents_file(file_path)
    
    @classmethod
    def save_in_file(cls, file_path, data, client_type=None, client=None):
        client = client if client is not None else cls.get_client(client_type)
        client.save_in_file(file_path, data)
    
    @classmethod
    def update_file(cls, file_path, data, update_type=UpdateType.AddOrUpdate, client_type=None, client=None, with_backup=True, backup_extension='.bak'):
        client = client if client is not None else cls.get_client(client_type)
        client.update_file(file_path, data, update_type, with_backup, backup_extension)
    
    @classmethod
    def restore_file(cls, file_path, client_type=None, client=None, backup_extension='.bak'):
        client = client if client is not None else cls.get_client(client_type)
        client.restore_file(file_path, backup_extension)
    
    
    # Manage YAML strings
    
    @classmethod
    def load_string(cls, text, client_type=None, client=None):
        client = client if client is not None else cls.get_client(client_type)
        return client.load_string(text)
    
    @classmethod
    def load_multiple_documents_string(cls, text, client_type=None, client=None):
        client = client if client is not None else cls.get_client(client_type)
        return client.load_multiple_documents_string(text)
    
    @classmethod
    def save_in_string(cls, data, client_type=None, client=None):
        client = client if client is not None else cls.get_client(client_type)
        return client.save_in_string(data)
    
    @classmethod
    def update_string(cls, text, data, update_type=UpdateType.AddOrUpdate, client_type=None, client=None):
        client = client if client is not None else cls.get_client(client_type)
        return client.update_string(text, data, update_type)





# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

# This file is a helper to build scripts using HolAdo solution, with CSV files as input.
# Usually, it is copied and adapted for specific usage.


import logging
import json
from holado_helper.script.behave_action_script import BehaveActionScript, BehaveActionScriptFile
from holado_helper.script.job import Job

logger = logging.getLogger(__name__)


class JsonMultipleActionScriptFile(BehaveActionScriptFile):
    def __init__(self):
        super().__init__(file_ext='.json')

    def read_jobs(self, file_path, action_script):
        converter = self.get_job_json_converter(action_script)
        
        with open(file_path, "r") as fin:
            file_json = json.load(fin)
        
        res = []
        for job_json in file_json:
            job = converter.from_json_to(job_json, Job)
            res.append(job)
        
        return res
    
    def write_jobs(self, file_path, jobs, action_script):
        raise NotImplementedError()
    
    def get_job_json_converter(self, action_script):
        from holado_helper.script.job import JobJsonConverter
        return JobJsonConverter(action_script)


class JsonMultipleActionScript(BehaveActionScript):
    def __init__(self, work_dir_path=None, actor=None, action_registry=None, script_file=None):
        if script_file is None:
            script_file = JsonMultipleActionScriptFile()
        super().__init__(work_dir_path=work_dir_path, with_action_parameter=False, actor=actor, action_registry=action_registry, script_file=script_file)
    
    def get_description(self, script_command, short_description):
        return """{short_description}
        
        Usage:
            1. Generate a template file with parameter '--output-template' as in example 1.
                It is possible to specify the name of the output file with parameter '-o'.
            2. Copy and modify the template file with desired actions. 
            3. Process actions described in file with parameter '-i'.
            4. See result in output file, if parameter '-o' is specified.
        
        In template files, a default value is set for each possible field. 
        For fields with enum values, the default value is a combination of each possible enum value separated by a '|', but only one value is possible.
        
        
        In input and output files, the field 'STATUS' describes the status of execution of each action:
            - 'TODO' (or ''): the action is to process (in output file, it means that execution was stopped on an error with a previous action)
            - 'SUCCESS': the action was in success
            - 'ERROR': the action was in error
            - others: the action is not processed, you can specify your own status to skip this action
        In template files, this column is automatically initialized with status 'TODO'.
        Any output file can be used as input file, the lines that are in status 'ERROR' or 'TODO' will be processed again.
        
        Example 1: Output template
                
                {script_command} --output-template -o actions.json
                
        Example 2: Process actions with given input file
                
                {script_command} -i actions.json
                
        """.format(short_description=short_description, script_command=script_command)
    
    def build_template_jobs(self):
        actions = []
        for action in self.action_registry.get_action_names():
            params_default_values = self.action_registry.get_action_parameters_default_values(action)
            params = {pdv[0]:pdv[1] for pdv in params_default_values}
            # Note: for templates, add static parameters is not needed
            act_info = self.new_action_info(action, params=params)
            actions.append(act_info)
        
        job = Job(actions=actions, in_parallel=False)
        
        return [job]






from builtins import super
from holado.common.context.session_context import SessionContext
import os
import logging

logger = logging.getLogger(__name__)


class TestHoladoSessionContext(SessionContext):
    
    def __init__(self):
        super().__init__("TestHoladoSessionContext")
    
    def configure(self, session_kwargs=None):
        logger.info("Configuring TestHoladoSessionContext")
        
        super().configure(session_kwargs)
        
        # Override default registered modules
        pass
        
        # Register new modules
        pass
    
    
    def initialize(self, session_kwargs=None):
        if session_kwargs is None:
            session_kwargs = {}
        raise_if_not_exist = session_kwargs.get("raise_if_not_exist", True)
        do_import = session_kwargs.get("import_compiled_proto", True)
        
        # Call default initialization
        super().initialize(session_kwargs)
        
        # Import compiled Protobuf and gRPC packages
        if do_import:
            here = os.path.abspath(os.path.dirname(__file__))
            proto_gene_path = os.path.join(here, "resources", "proto", "generated")
            
            SessionContext.instance().protobuf_messages.import_all_compiled_proto(os.path.join(proto_gene_path, "protobuf"), raise_if_not_exist=raise_if_not_exist)
        
        
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import copy
from holado_core.common.finders.else_finder import ElseFinder
from holado_core.common.exceptions.technical_exception import TechnicalException


class InspectParameters(object):
    """ Inspect parameters
    """
    
    __instance_default = None
    __instance_default_without_raise = None
    
    def __init__(self, raise_no_such_finder=True, finder_types=None, result_finder_type=None):
        """
        Constructor
        @param raise_no_such_finder If True, raise NoSuchElementException, else return None, when no element is found.
        @param finder_types Finder types to use
        @param result_finder_type Type of finder result when multiple finders are possible
        """
        self.raise_no_such_finder = raise_no_such_finder
        self.finder_types = [] if finder_types is None else finder_types
        self.result_finder_type = ElseFinder if result_finder_type is None else result_finder_type
    
    def get_module_parameters(self, inspect_parameters):
        """
        @param inspect_parameters Current inspect parameters
        @return Module inspect parameters updated with current inspect parameter
        """
        return self.with_result_finder_type(inspect_parameters.result_finder_type) \
            .with_raise(inspect_parameters.raise_no_such_finder)
    
    def set_raise(self, raise_exception):
        """
        @param raise_exception Whether raise exceptions or not
        """
        self.raise_no_such_finder = raise_exception
    
    def with_raise(self, raise_exception=True):
        """
        Note: if raise booleans are same, self instance is returned, else a new one is returned
        @param raise_exception Whether raise exceptions or not
        @return Same parameters but with raise
        """
        if self.raise_no_such_finder == raise_exception:
            return self
        else:
            res = copy.deepcopy(self)
            res.raise_no_such_finder = raise_exception
            return res
        
    def without_raise(self):
        """
        Note: if raise booleans are already False, self instance is returned, else a new one is returned
        @return Same parameters but without raise
        """
        return self.with_raise(False)
    
    def add_finder_type(self, finder_type):
        """
        @param finder_type Finder type to add
        """
        if finder_type in self.finder_types:
            raise TechnicalException(f"Finder type '{finder_type}' already exists")
        self.finder_types.append(finder_type)
    
    def add_finder_types(self, finder_types):
        """
        @param finder_types Finder types to add
        """
        for finder_type in finder_types:
            self.add_finder_type(finder_type)
    
    def remove_finder_type(self, finder_type):
        """
        @param finder_type Finder type to remove
        """
        if finder_type not in self.finder_types:
            raise TechnicalException(f"Finder type '{finder_type}' doesn't exist")
        self.finder_types.remove(finder_type)
    
    def remove_finder_types(self, finder_types):
        """
        @param finder_types Finder types to remove
        """
        for finder_type in finder_types:
            self.remove_finder_type(finder_type)
    
    def remove_all_finder_types(self):
        """
        Remove all finder types
        """
        self.finder_types.clear()
    
    def with_finder_type(self, finder_type):
        """
        Note: if wanted finder type is already set, this instance is returned, else a new one is returned
        @param finder_type Finder type
        @return Same parameters but with given finder type
        """
        if finder_type in self.finder_types:
            return self
        else:
            res = copy.deepcopy(self)
            res.finder_types.append(finder_type)
            return res
        
    def with_finder_types(self, finder_types):
        """
        Note: if wanted finder types are already set, this instance is returned, else a new one is returned
        @param finder_types Finder types
        @return Same parameters but with given finder types
        """
        res = self
        for finder_type in finder_types:
            res = res.with_find_type(finder_type)
        return res
        
    def set_result_finder_type(self, result_finder_type):
        """
        @param result_finder_type Result finder type
        """
        self.result_finder_type = result_finder_type
    
    def with_result_finder_type(self, result_finder_type):
        """
        Note: if given result finder class is already set, this instance is returned, else a new one is returned
        @param result_finder_type Result finder class
        @return Same parameters but with given result finder class
        """
        if self.result_finder_type == result_finder_type:
            return self
        else:
            res = copy.deepcopy(self)
            res.result_finder_type = result_finder_type
            return res
        
    @staticmethod
    def default(raise_exception=True):
        """
        @param raise_exception Whether raise exceptions or not
        @return Default inspect parameters with given raiseability
        """
        if raise_exception:
            return InspectParameters.default_with_raise()
        else:
            return InspectParameters.default_without_raise()
    
    @staticmethod
    def default_with_raise():
        """
        @return Default find parameters with raises
        """
        if InspectParameters.__instance_default is None:
            InspectParameters.__instance_default = InspectParameters()
        return InspectParameters.__instance_default
    
    @staticmethod
    def default_without_raise():
        """
        @return Default find parameters without any raise
        """
        if InspectParameters.__instance_default_without_raise is None:
            InspectParameters.__instance_default_without_raise = InspectParameters(raise_no_such_finder=False)
        return InspectParameters.__instance_default_without_raise
    

    
    
    


#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.handlers.element_holder import ElementHolder
from holado_core.common.exceptions.technical_exception import TechnicalException
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.common.alert import Alert
from selenium.webdriver.remote.webdriver import WebDriver
from holado_python.standard_library.typing import Typing


class SeleniumHolder(ElementHolder):
    """ Information on a Selenium element.
    """
    
    def __init__(self, parent, element, selenium_driver, description=None):
        """
        @param parent Parent
        @param element Element instance
        @param selenium_driver Selenium driver
        @param description Element description
        """
        super().__init__(parent, element, description)
        self.__selenium_driver = selenium_driver

    @property
    def driver(self):
        """
        @return Selenium driver
        """
        return self.__selenium_driver
    
    @property
    def is_alert(self):
        """
        @return If is an alert
        """
        return isinstance(self.element, Alert)
    
    @property
    def is_web_driver(self):
        """
        @return If is a web driver
        """
        return isinstance(self.element, WebDriver)
    
    @property
    def is_web_element(self):
        """
        @return If is a web element
        """
        return isinstance(self.element, WebElement) and not self.is_popup
    
    @property
    def is_popup(self):
        """
        @return If is a popup
        """
        return False

    @ElementHolder.complete_description_and_details.getter  # @UndefinedVariable
    def complete_description_and_details(self):
        if self.is_alert:
            return f"{self.complete_description} (element: {{ALERT}})"
        elif self.is_web_driver:
            return f"{self.complete_description} (element: {{PAGE}})"
        elif self.is_web_element:
            return f"{self.complete_description} (element: {self.driver.internal_api.get_element_description(self.element)})"
        else:
            return super().complete_description_and_details

    def get_holder_for(self, element, description):
        """
        @param element New element instance.
        @param description New element description.
        @return Element holder for given element instance
        """
        if isinstance(element, WebElement):
            return SeleniumHolder(self, element, self.driver, description)
        else:
            raise TechnicalException(f"Unexpected element type {Typing.get_object_class_fullname(element)}")
    
    def get_popup_holder_for(self, element, title, description):
        """
        @param element Popup element
        @param title Popup title
        @param description Popup description
        @return Popup holder for given popup element
        """
        if isinstance(element, WebElement):
            return PopupHolder(self, element, self.driver, title, description)
        else:
            raise TechnicalException(f"Unexpected element type {Typing.get_object_class_fullname(element)}")
    
    
    
class PopupHolder(SeleniumHolder):
    """ Information on a popup element.
    """
    
    def __init__(self, parent, element, selenium_driver, title=None, description=None):
        """
        @param parent Parent
        @param element Popup element instance
        @param selenium_driver Selenium driver
        @param title Popup title (None means no title)
        @param description Element description
        """
        super().__init__(parent, element, selenium_driver, description)
        self.__title = title
        
    @SeleniumHolder.is_popup.getter
    def is_popup(self):
        return True

    @property
    def title(self):
        """
        @return Popup title
        """
        return self.__title
    
    @ElementHolder.complete_description_and_details.getter  # @UndefinedVariable
    def complete_description_and_details(self):
        if self.is_popup:
            if self.title is not None:
                return f"{self.complete_description} (popup '{self.title}' element: {self.driver.internal_api.get_element_description(self.element)})"
            else:
                return f"{self.complete_description} (popup element: {self.driver.internal_api.get_element_description(self.element)})"
        else:
            return super().complete_description_and_details
    
    
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado.common.handlers.object import DeleteableObject
from holado_multitask.multithreading.functionthreaded import FunctionThreaded
from holado_python.common.tools.datetime import DateTime
import json
from holado_json.filesystem.stream_json_file import StreamJSONFile
import re
from holado.common.context.session_context import SessionContext
import os
from datetime import datetime

logger = logging.getLogger(__name__)




class DockerContainerLogsFollower(DeleteableObject):
    """Generic follower of logs of a container
    """
    def __init__(self, docker_container):
        super.__init__(self, docker_container.container.name)
        self.__docker_container = docker_container
        self.__logs = []
        self.__logs_stream = None
    
    def _delete_object(self):
        self.close()
    
    def close(self):
        if self.__logs_stream is not None:
            self.__logs_stream.close()
            self.__thread.join()
            
            self.__logs_stream = None
            self.__thread = None
    
    def follow_logs(self, stdout=True, stderr=True, tail='all', since=None):
        """Follow logs from Docker Engine
        """
        if isinstance(since, str):
            since = DateTime.str_2_datetime(since)
        
        if self.__logs_stream is not None:
            raise TechnicalException(f"Logs of container '{self.name}' are already followed")
        self.__logs_stream = self.__docker_container.container.logs(stdout=stdout, stderr=stderr, tail=tail, since=since, stream=True, timestamps=True)
        
        self.__thread = FunctionThreaded(self.__read_stream, name=f"follow logs of container '{self.__docker_container.container.name}'", register_thread=False)
        self.__thread.interrupt_function = self.__logs_stream.close
        self.__thread.start()
    
    def __read_stream(self):
        for log in self.__logs_stream:
            self.__logs.append( (log['time'], log['steam'] == 'stdout', self.__docker_container._get_log_from_log_bytes(log['log'])) )
    
    def reset(self):
        self.__logs.clear()
    
    def get_logs(self, stdout=True, stderr=True, timestamps=False, tail='all', since=None, until=None):
        if since is not None and isinstance(since, datetime):
            since = DateTime.datetime_2_str(since)
        if until is not None and isinstance(until, datetime):
            until = DateTime.datetime_2_str(until)
        
        res = []
        for log in self.__logs:
            if log[1] and not stdout or not log[1] and not stderr:
                continue
            if since is not None and log[0] < since:
                continue
            if until is not None and log[0] >= until:
                continue
            
            if timestamps:
                res.append( (log[0], log[2]) )
            else:
                res.append(log[2])
        
        if tail != 'all':
            if not isinstance(tail, int) or tail < 0:
                raise TechnicalException(f"Argument 'tail' can be 'all' or a positive integer")
            res = res[-tail:]
        
        return res




class DockerContainerLogsFormatter(object):
    """Base class for formatters of container logs.
    """
    def format_logs(self, logs, with_timestamps=False):
        raise NotImplementedError()

class JsonDockerContainerLogsFormatter(DockerContainerLogsFormatter):
    """Formatter of container logs in JSON format.
    """
    def __init__(self, timestamps_key='time'):
        self.__timestamps_key = timestamps_key
    
    def format_logs(self, logs, with_timestamps=False):
        res = []
        
        for log in logs:
            ts = None
            if with_timestamps:
                ts, log = log[0], log[1]
            
            res_log = {}
            
            # Begin by timestamp so that it is the first dict element
            if ts is not None:
                res_log[self.__timestamps_key] = ts
            
            # Add log as JSON format
            try:
                log_dict = json.loads(log)
            except:
                # In some circumstances (ex: container is in panic), logs can be in another format even if configured in json format
                log_dict = {'log':log}
            res_log.update(log_dict)
            
            res.append(res_log)
        
        return res




class DockerContainerLogsSaver(object):
    """Base class for savers of docker container logs.
    """
    def save_logs(self, docker_container, file_path, stdout=True, stderr=True, timestamps=True, tail='all', since=None, until=None):
        """ Save logs of a container.
        @return: number of logs saved
        """
        raise NotImplementedError()

class JsonDockerContainerLogsSaver(DockerContainerLogsSaver):
    """Saver of docker container logs in a stream JSON file.
    """
    def __init__(self, timestamps_key='time', **dumps_kwargs):
        self.__timestamps_key = timestamps_key
        self.__dumps_kwargs = dumps_kwargs
        
    def save_logs(self, docker_container, file_path, stdout=True, stderr=True, timestamps=True, tail='all', since=None, until=None):
        res = None
        SessionContext.instance().path_manager.makedirs(file_path)
        
        logs = docker_container.get_logs(stdout=stdout, stderr=stderr, timestamps=timestamps, tail=tail, since=since, until=until, 
                                         formatter=JsonDockerContainerLogsFormatter(timestamps_key=self.__timestamps_key))
        if logs:
            with StreamJSONFile(file_path, mode='wt') as fout:
                fout.write_elements_json_object_list(logs)
            res = len(logs)
            logger.info(f"Saved {res} logs for container '{docker_container.container.name}' in file '{file_path}'")
        
        return res


class DockerContainersLogsSaver():
    """Saver of logs for a group of containers
    """
    def __init__(self, docker_client, container_logs_saver=JsonDockerContainerLogsSaver(), include_patterns=None, exclude_patterns=None):
        self.__docker_client = docker_client
        self.__container_logs_saver = container_logs_saver
        self.__include_patterns = [ip if isinstance(ip, re.Pattern) else re.compile(ip) for ip in include_patterns] if include_patterns is not None else None
        self.__exclude_patterns = [ep if isinstance(ep, re.Pattern) else re.compile(ep) for ep in exclude_patterns] if exclude_patterns is not None else None
        
    def save_containers_logs(self, destination_path, stdout=True, stderr=True, timestamps=True, tail='all', since=None, until=None):
        """ Save containers logs
        @return: number of containers for which logs have been saved
        """
        res = 0
        
        names = self.__docker_client.get_container_names(in_list=True, all_=True, sparse=True, include_patterns=self.__include_patterns, exclude_patterns=self.__exclude_patterns)
        for name in names:
            container = self.__docker_client.get_container(name, all_=True, reset_if_removed=False)
            file_path = os.path.join(destination_path, f"{name}.log")
            nb_logs = self.__container_logs_saver.save_logs(container, file_path, stdout=stdout, stderr=stderr, timestamps=timestamps, tail=tail, since=since, until=until)
            
            if nb_logs is not None:
                res += 1
        
        return res




# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


import logging
from holado_core.common.tools.tools import Tools
from holado_core.common.tables.table_with_header import TableWithHeader
from holado_value.common.tables.comparators.table_2_value_table_comparator import Table2ValueTable_Comparator
from holado_core.common.tables.table_manager import TableManager
from holado_value.common.tables.value_table_manager import ValueTableManager
from holado.common.context.session_context import SessionContext
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_test.behave.behave import *  # @UnusedWildImport
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from holado_test.scenario.step_tools import StepTools
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


# Configure module to use internally
try:
    from suds.sudsobject import Printer, Object, Facade  # @UnresolvedImport @UnusedImport
    import suds.client  # @UnresolvedImport
    from suds.cache import NoCache  # @UnresolvedImport
    with_suds = True
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug("Module 'suds' will be used in web service steps")
except Exception as exc:
    with_suds_exception = exc
    with_suds = False

if not with_suds:
    try:
        import zeep  # @UnresolvedImport
        from zeep.transports import Transport  # @UnresolvedImport
        with_zeep = True
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Module 'zeep' will be used in web service steps")
    except Exception as exc:
        with_zeep_exception = exc
        with_zeep = False

if not with_suds and not with_zeep:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"Web service module is not available. No package was found to manage web service steps (managed packages: suds, zeep).\nInitialization failed with suds on error: {with_suds_exception}\nInitialization failed with zeep on error: {with_zeep_exception}")


def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_text_interpreter():
    return __get_scenario_context().get_text_interpreter()

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()

@Given(r"open web service client for WSDL (?P<wsdl_file>{Str})")
def step_impl(context, wsdl_file):
    wsdl_file = StepTools.evaluate_scenario_parameter(wsdl_file)
    
    if with_suds:
        context.ws_client = suds.client.Client(wsdl_file, cache=NoCache())
    elif with_zeep:
        transport = Transport(cache=None)
        context.ws_client = zeep.Client(wsdl_file, transport=transport)
    else:
        raise TechnicalException("No module was found to manage web service steps (managed modules: suds, zeep)")

@Given(r"open (?P<ws_name>{Str}) web service client")
def step_impl(context, ws_name):  # @DuplicatedSignature
    ws_name = StepTools.evaluate_scenario_parameter(ws_name)
    
    wsdl_file = 'http://10.0.3.247/{}/?wsdl'.format(ws_name)
    step = u"Given open web service client for WSDL '{}'".format(wsdl_file)
    execute_steps(step)

@Given(r'close web service client')
def step_impl(context):  # @DuplicatedSignature
    del context.ws_client

@When(r"(?P<var_name>{Variable}) = call web service (?P<method_name>{Str}) with parameters")
def step_impl(context, var_name, method_name):  # @DuplicatedSignature
    var_name = StepTools.evaluate_variable_name(var_name)
    method_name = StepTools.evaluate_scenario_parameter(method_name)
    table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    parameters = ValueTableConverter.convert_name_value_table_2_dict(table)
    logger.info("call web service '{}' with parameters [{}]".format(method_name, parameters))
    
    # Verify web service method
    if not hasattr(context, 'ws_client') or context.ws_client is None:
        raise FunctionalException("No web service is opened")
    try:
        method = getattr(context.ws_client.service, method_name)
    except:
        raise FunctionalException("Method '{}' doesn't exist in web service".format(method_name))
    
    # Call web service method
    result = method(**parameters)

    # Store result in variable
    logger.info("result of web service '{}': [{}]".format(method_name, result))
    __get_variable_manager().register_variable(var_name, result)


@Given(r"(?P<var_name>{Variable}) = new (?P<type_name>{Str}) from web service")
def step_impl(context, var_name, type_name):  # @DuplicatedSignature
    var_name = StepTools.evaluate_variable_name(var_name)
    type_name = StepTools.evaluate_string_parameter(type_name)
    # Verify web service
    if not hasattr(context, 'ws_client') or context.ws_client is None:
        raise FunctionalException("No web service is opened")

    table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    
    # Create object instance
    if with_suds:
        obj = context.ws_client.factory.create(type_name)
    elif with_zeep:
        obj = getattr(context.ws_client.type_factory('asgard.api_reseaux.tns'), type_name)()
    else:
        raise TechnicalException("No module was found to manage web service steps (managed modules: suds, zeep)")
    
    # Fill object
    if type_name == "stringArray":
        # Verify table structure
        ValueTableManager.verify_table_is_value_table(table)
        
        # Fill stringArray
        for row in table.rows:
            obj.string.append(row.get_cell(0).value)
    else:
        ValueTableManager.set_object_attributes_according_name_value_table(obj, table)
    
    # Store in variable
    __get_variable_manager().register_variable(var_name, obj)

@Then(r"(?P<object_string>{Str}) is (?P<type_name>{Str}) from web service")
def step_impl(context, object_string, type_name):  # @DuplicatedSignature
    obj = StepTools.evaluate_scenario_parameter(object_string)
    type_name = StepTools.evaluate_string_parameter(type_name)
    table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    
    # Verify web service
    if not hasattr(context, 'ws_client') or context.ws_client is None:
        raise FunctionalException("No web service is opened")
    
    if with_suds:
        ws_obj = context.ws_client.factory.create(type_name)
    elif with_zeep:
        ws_obj = getattr(context.ws_client.type_factory('asgard.api_reseaux.tns'), type_name)()
    else:
        raise TechnicalException("No module was found to manage web service steps (managed modules: suds, zeep)")
        
    if isinstance(obj, Typing.get_object_class(ws_obj)):
        if type_name == "stringArray":
            # Convert stringArray to table
            obj_table = TableWithHeader()
            obj_table.header.add_cells_from_contents(["Value"])
            for obj_str in obj.string:
                obj_table.add_row(cells_content=[obj_str])
            
            # Compare tables
            comparator = Table2ValueTable_Comparator()
            comparator.equals(obj_table, table)
        else:
            # Convert object to table
            obj_table = TableManager.convert_object_attributes_2_name_value_table(obj)
            
            # Compare tables
            comparator = Table2ValueTable_Comparator()
            comparator.equals(obj_table, table)
    else:
        raise FunctionalException("Object is of type '{}' (expected type: '{}' ; type_name: '{}')".format(Typing.get_object_class_fullname(obj), Typing.get_object_class_fullname(ws_obj), type_name))





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.tools.tools import Tools
from holado.holado_config import Config
from holado_core.common.exceptions.element_exception import NoSuchElementException
from holado_core.common.finders.element_finder import ElementFinder
from holado_core.common.finders.tools.find_parameters import FindParameters
from holado_core.common.finders.tools.find_context import ContainerFindContext

logger = logging.getLogger(__name__)


class ThenFinder(ElementFinder):
    """ Finder that search successively in a list of self.__finders.
    Search is done from first added to last added finder, using output of previous finder as input of next finder.
    """
    
    def __init__(self, description=None):
        super().__init__(description)
        self.__self.__finders = []
    
    def set_next_finder(self, finder, find_type=None, update_root_container=None):
        """
        Append given finder in then succession.
        @param finder A finder.
        @param find_type Find method type.
        """
        if find_type is not None:
            finder.find_type = find_type
        if update_root_container is not None:
            finder.find_updater.set_update_root(update_root_container)
            
        # Only first finder can unspecify the find type
        if len(self.__self.__finders) > 0:
            finder.verify_find_type_is_defined("Except first finder, all others must specify find type.")
            
        self.__self.__finders.append(finder)
    
    def _find_all_container(self, find_context:ContainerFindContext, find_parameters:FindParameters):
        res = None
        stop = False
        sub_parameters = find_parameters.get_next_level()
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"{self._get_indent_string_level(find_parameters)}[ThenFinder({self.element_description})._find_all_container] -> begin in [{find_context.get_input_description()}]")
        
        # Prepare analyze time spent
        if find_parameters.analyze_time_spent:
            start = Tools.timer_s()
        
        finder = self.__self.__finders[0]
        try:
            # Process first finder
            res = finder._find_all_container(find_context, sub_parameters)
        except NoSuchElementException as exc:
            if (1 < len(self.__self.__finders) and find_parameters.raise_intermediate_no_such_element
                    or 1 == len(self.__self.__finders) and find_parameters.raise_no_such_element):
                Tools.raise_same_exception_type(exc, f"[ThenFinder({self.element_description})[0]: {finder.element_description}] {exc.message}")
            else:
                stop = True
        except Exception as exc:
            Tools.raise_same_exception_type(exc, f"[ThenFinder({self.element_description})[0]: {finder.element_description}] {exc.message}")
        
        # Process next self.__finders
        if not stop and len(self.__self.__finders) > 1:
            res = self.__process_next_finder(res, 1, finder, find_context, find_parameters)
        
        # Analyze time spent
        if find_parameters.analyze_time_spent:
            duration = Tools.timer_s() - start
            if duration > Config.threshold_warn_time_spent_s:
                logger.warning(f"ThenFinder[{self.element_description}]._find_all_container end (took: {duration} s)     -> {len(res)} elements")
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            if len(res) == 0:
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"{self._get_indent_string_level(find_parameters)}[ThenFinder({self.element_description})._find_all_container] -> return 0 candidate")
            else:
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"{self._get_indent_string_level(find_parameters)}[ThenFinder({self.element_description})._find_all_container] -> return {len(res)} candidates:\n{self._represent_candidates_output(res, 4)}")
        
        if len(res) == 0 and find_parameters.raise_no_such_element:
            raise NoSuchElementException(f"[ThenFinder({self.element_description})._find_all_container] Unable to find any element")
        return res
    
    def __process_next_finder(self, candidates, finder_index, previous_finder, find_context, find_parameters):
        res = []
        cur_container = None
        candidates_for_next_finder = None
        sub_find_parameters = find_parameters.get_next_level()
        
        # Manage intermediate result
        if len(candidates) == 0:
            if find_parameters.raise_intermediate_no_such_element:
                raise NoSuchElementException(f"[ThenFinder({self.element_description})[{finder_index-1}]] Unable to find intermediate {previous_finder.element_description}")
            # Else process next finders in case of finder managing no result
        
        # Get finder to use
        finder = self.__finders.get(finder_index)
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            prefix_message = f"{self._get_indent_string_level(find_parameters)}[ThenFinder({self.element_description}).__process_next_finder[{finder_index}]: {finder.element_description}]"
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"{prefix_message} -> begin with {len(candidates)} candidates:\n{self._represent_candidates_output(res, 4)}")
        
        # Process finder on each candidate
        for i, cur_container in enumerate(candidates):
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"{prefix_message}     -> candidate {i}: {self._represent_candidate_output(cur_container)}")
            
            # Process finder
            try:
                candidates_for_next_finder = finder.find_all(container=cur_container, find_context=find_context, find_parameters=sub_find_parameters)
            except NoSuchElementException:
                # Nothing, continue with another candidate
                continue
            except Exception as exc:
                Tools.raise_same_exception_type(exc, f"[ThenFinder({self.element_description})[{finder_index}]: {finder.element_description} - on candidate {i+1}/{len(candidates)}] {exc.message}")
                
            # Process next finder if existing
            if finder_index + 1 < len(self.__finders):
                try:
                    self._add_candidates(res, self.__process_next_finder(candidates_for_next_finder, finder_index + 1, finder, find_context, find_parameters))
                except NoSuchElementException:
                    # Nothing, continue with another candidate
                    continue
                
            else:
                self._add_candidates(res, candidates_for_next_finder)
                if Tools.do_log(logger, logging.TRACE) and len(candidates_for_next_finder) > 0:  # @UndefinedVariable
                    logger.trace(f"{prefix_message}     -> candidate {i} => {len(candidates_for_next_finder)} elements found")
            
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            if len(res) == 0:
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"{prefix_message} -> return 0 candidate")
            else:
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"{prefix_message} -> return {len(res)} candidates:\n{self._represent_candidates_output(res, 4)}")

        return res
    
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_test.scenario.step_tools import StepTools
from holado_value.common.tables.value_table import ValueTable
from holado_scripting.common.tools.evaluate_parameters import EvaluateParameters
import behave
from holado_core.common.tools.tools import Tools



logger = logging.getLogger(__name__)


class BehaveStepTools(StepTools):
    """
    Gives usefull tools for steps files.
    
    Warning: In some cases, use of both step matcher "parse" and a parameter type 
             will have mistake behaviours.
             For example, the result can be that steps of other files are missing at execution for behave
             whereas they are well found and imported in behave.
             If this happens, use step matcher "re"
    """
    
    ############################################################################
    ## Manage step definitions



    
    DEFAULT_STEP_MATCHER = "re"
    
    @classmethod    
    def use_step_matcher(cls, name):
        behave.use_step_matcher(name)
    
    @classmethod    
    def use_step_matcher_default(cls):
        cls.use_step_matcher(cls.DEFAULT_STEP_MATCHER)
    
    # @classmethod
    # def import_types(cls, module, prefix="SRE_", do_upper=False):
    #     res = []
    #     for type_name in cls.__registered_types:
    #         var_name = f"{prefix}{type_name}"
    #         if do_upper:
    #             var_name = var_name.upper()
    #         if not hasattr(module, var_name):
    #             setattr(module, var_name, cls.get_registered_type_pattern(type_name))
    #             res.append(var_name)
    #     logger.debug(f"Imported step parameter types in module '{module.__name__}' as variables: {res}")
    #     return res
    
    @classmethod    
    def register_type(cls, type_name, pattern, eval_func, **eval_func_kwargs):
        super().register_type(type_name, pattern, eval_func, **eval_func_kwargs)
        
        # Register type in parse step matcher
        func = cls.get_registered_type_function(type_name)
        cls.use_step_matcher('parse')
        behave.register_type(**{type_name:func})
        cls.use_step_matcher_default()
    
    
    
    ############################################################################
    ## Manage step multiline text
    
    @classmethod    
    def get_step_multiline_text(cls, context, eval_params=EvaluateParameters.default_without_raise(), raise_exception_if_none=True, log_level=logging.DEBUG):
        """
        Get the step multiline text interpreted.
        Note: parameter context can be the step context or the step object.
        """
        res = None
        if hasattr(context, "text") and context.text is not None:
            res = context.text.strip()
            if eval_params.do_interpret:
                res = cls._get_text_interpreter().interpret(res, eval_params=eval_params, log_level=log_level)
            
        if res is None and raise_exception_if_none:
            raise FunctionalException(f"Multiline text is required under this step\n  context: {context}")
        
        return res
        
    
    ############################################################################
    ## Manage step table
    
    @classmethod
    def _is_scenario_step_table(cls, table):
        return isinstance(table, behave.model.Table)
    
    @classmethod
    def get_step_table(cls, context, do_eval_once=True):
        """
        Get the step table interpreted.
        Note: parameter context can be the step context or the step object.
        """
        if hasattr(context, "table") and context.table is not None:
            if cls.has_table_header(context.table):
                res = cls.convert_step_table_2_value_table_with_header(context.table, do_eval_once=do_eval_once)
            else:
                res = cls.convert_step_table_2_value_table(context.table, do_eval_once=do_eval_once)
            return res
        else:
            return None
        
    @classmethod        
    def _get_table_header(cls, table):
        return table.headings

    @classmethod
    def _get_table_rows(cls, table):
        return table
 
    @classmethod
    def convert_step_table_2_value_table(cls, table, do_eval_once=True):
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Converting step table to scenario table (step table = {table})")

        res = ValueTable()
        
        # Set header as row
        previous_row = None
        srow = cls._convert_step_table_row_2_value_table_row(cls._get_table_header(table), previous_row, do_eval_once=do_eval_once)
        res.add_row(row=srow)
        
        # Set body
        previous_row = srow
        for row in cls._get_table_rows(table):
            srow = cls._convert_step_table_row_2_value_table_row(row, previous_row, do_eval_once=do_eval_once)
            res.add_row(row=srow)
            previous_row = srow
            
        # Manage one cell table referencing a variable containing a table
        res = cls._extract_inner_table_if_present(res, do_eval_once=do_eval_once)
            
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Converting step table to scenario table (step table = {table}) => {res}")
        return res
        
        
        



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: account.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\raccount.proto\x12\x07\x61\x63\x63ount\x1a\x1bgoogle/protobuf/empty.proto\"C\n\x04User\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x10\n\x08username\x18\x02 \x01(\t\x12\r\n\x05\x65mail\x18\x03 \x01(\t\x12\x0e\n\x06groups\x18\x04 \x03(\x05\"\x11\n\x0fUserListRequest\"!\n\x13UserRetrieveRequest\x12\n\n\x02id\x18\x01 \x01(\x05\x32\x88\x02\n\x0eUserController\x12\x33\n\x04List\x12\x18.account.UserListRequest\x1a\r.account.User\"\x00\x30\x01\x12(\n\x06\x43reate\x12\r.account.User\x1a\r.account.User\"\x00\x12\x39\n\x08Retrieve\x12\x1c.account.UserRetrieveRequest\x1a\r.account.User\"\x00\x12(\n\x06Update\x12\r.account.User\x1a\r.account.User\"\x00\x12\x32\n\x07\x44\x65stroy\x12\r.account.User\x1a\x16.google.protobuf.Empty\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'account_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_USER']._serialized_start=55
  _globals['_USER']._serialized_end=122
  _globals['_USERLISTREQUEST']._serialized_start=124
  _globals['_USERLISTREQUEST']._serialized_end=141
  _globals['_USERRETRIEVEREQUEST']._serialized_start=143
  _globals['_USERRETRIEVEREQUEST']._serialized_end=176
  _globals['_USERCONTROLLER']._serialized_start=179
  _globals['_USERCONTROLLER']._serialized_end=443
# @@protoc_insertion_point(module_scope)
